<?php
/**
 * Shlib - programming library
 *
 * @author      Yannick Gaultier
 * @copyright   (c) Yannick Gaultier 2016
 * @package     shlib
 * @license     http://www.gnu.org/copyleft/gpl.html GNU/GPL
 * @version     0.3.1.603
 * @date                2017-02-17
 */

defined('_JEXEC') or die;

/**
 * Simple external authentication system
 *
 */
class ShlSystem_Auth
{
	/**
	 * Sign an outgoing request with (our) standard headers
	 *
	 * Note that they query should be used to send the request without being modified or added to, at least if the signature is being checked
	 * by the receiving end.
	 *
	 * Also, the query variables are alphabetically sorted on the array key (ie the query variable name)
	 * prior to signature being computed so as to normalize the input and insure repeatability on both ends
	 *
	 * @param array  $query Key/value array of query variables
	 * @param string $authKey A secret key shared between emitter and receiver
	 *
	 * @return Object  'query' => built query string (ie p1=123&p2=456...), 'urlEncodedQuery' => same as query but url encoded, 'headers' => key/value array of headers to be sent
	 */
	public static function signRequest(
		$query,
		$authKey
	)
	{
		$request = new stdClass();
		$headers = array(
			'X-WBLR-AUTH-TS' => time(),
			'X-WBLR-AUTH-ID' => JString::trim($authKey),
			'X-WBLR-AUTH-TOKEN' => sha1(__FILE__ . microtime())
		);

		// normalize
		ksort($query);

		// build the request, to be signed
		$queryString = array();
		$queryUrlEncoded = array();
		foreach ($query as $key => $value)
		{
			$queryString[] = $key . '=' . $value;
			$queryUrlEncoded[] = $key . '=' . urlencode($value);
		}
		$queryString = implode('&', $queryString);
		$queryUrlEncoded = implode('&', $queryUrlEncoded);

		$base = $headers['X-WBLR-AUTH-TS']
			. $headers['X-WBLR-AUTH-ID']
			. $headers['X-WBLR-AUTH-TOKEN']
			. $queryString;
		$headers['X-WBLR-AUTH-SIG'] = sha1($base);

		$request->query = $queryString;
		$request->urlEncodedQuery = $queryUrlEncoded;
		$request->headers = $headers;

		return $request;
	}

	/**
	 * Verify the integrity of an incoming request
	 *
	 * @param array $query Key/value array of query variables
	 * @param array $incomingHeaders
	 * @param bool  $urlDecodeBeforeVerify Whether query string should be urldecode-d before auth is verified
	 *
	 * @return Object  'status' => HTTP status code, 'message' => Description of the response status
	 */
	public static function verifyRequest(
		$query,
		$incomingHeaders,
		$allowedTimeSkew,
		$urlDecodeBeforeVerify = true
	)
	{
		$verifiedRequest = new stdClass();
		$verifiedRequest->code = 200;
		$verifiedRequest->message = 'OK';

		$headers = array_merge(
			array(
				'X-WBLR-AUTH-TS' => '',
				'X-WBLR-AUTH-ID' => '',
				'X-WBLR-AUTH-TOKEN' => '',
				'X-WBLR-AUTH-SIG' => '',
			),
			$incomingHeaders
		);

		// prevent edge cases when values are not supplied
		if (
			empty($headers['X-WBLR-AUTH-TS'])
			||
			empty($headers['X-WBLR-AUTH-ID'])
			||
			empty($headers['X-WBLR-AUTH-TOKEN'])
			||
			empty($headers['X-WBLR-AUTH-SIG'])
		)
		{
			$verifiedRequest->message = 'Not authorized (invalid headers).';
			$verifiedRequest->code = 403;
		}

		if (!self::hasValidTimeSkew(
			$headers['X-WBLR-AUTH-TS'],
			$allowedTimeSkew
		)
		)
		{
			$verifiedRequest->message = 'Not authorized (invalid timestamp).';
			$verifiedRequest->code = 403;
		}

		if (!self::hasValidSignature($query, $headers, $urlDecodeBeforeVerify))
		{
			$verifiedRequest->message = 'Not authorized (invalid signature).';
			$verifiedRequest->code = 403;
		}

		return $verifiedRequest;
	}

	/**
	 * Check whether the request time stamp is older than a given threshold
	 *
	 * An allowedTimeStamp value of 0 disables the test
	 *
	 * @param int $requestTimeStamp
	 * @param int $allowedTimeSkew
	 *
	 * @return bool
	 */
	private static function hasValidTimeSkew($requestTimeStamp, $allowedTimeSkew)
	{
		$skew = time() - (int) $requestTimeStamp;
		if (!empty($allowedTimeSkew) && abs($skew) > $allowedTimeSkew)
		{
			return false;
		}

		return true;
	}

	/**
	 * Run signing method on query to verify it matches the passed signature
	 *
	 * @param array $request Key/value array of query variables
	 * @param array $headers
	 * @param bool  $urlDecodeBeforeVerify Whether query string should be urldecode-d before auth is verified
	 */
	private static function hasValidSignature($query, $headers, $urlDecodeBeforeVerify = false)
	{
		$base = $headers['X-WBLR-AUTH-TS'] . $headers['X-WBLR-AUTH-ID'] . $headers['X-WBLR-AUTH-TOKEN'];

		// sort query array by key, to normalize hash building
		ksort($query);

		// build up query string
		$queryString = array();
		foreach ($query as $key => $value)
		{
			$queryString[] = $key . '=' . ($urlDecodeBeforeVerify ? urldecode($value) : $value);
		}
		$base .= implode('&', $queryString);

		// now verify signature against the one passed in the request
		$computedSignature = sha1($base);
		if ($computedSignature != $headers['X-WBLR-AUTH-SIG'])
		{
			return false;
		}

		return true;
	}
}
